// init flash file system
void init_spiffs (void)
{
    ESP_LOGI(TAG, "Initializing SPIFFS");
    
    esp_vfs_spiffs_conf_t conf = {
      .base_path = "/spiffs",
      .partition_label = NULL,
      .max_files = 5,
      .format_if_mount_failed = true
    };
    
    // Use settings defined above to initialize and mount SPIFFS filesystem.
    // Note: esp_vfs_spiffs_register is an all-in-one convenience function.
    esp_err_t ret = esp_vfs_spiffs_register(&conf);

    if (ret != ESP_OK) 
        {
        if (ret == ESP_FAIL) 
            ESP_LOGE(TAG, "Failed to mount or format filesystem");
        else if (ret == ESP_ERR_NOT_FOUND)
            ESP_LOGE(TAG, "Failed to find SPIFFS partition");
        else
            ESP_LOGE(TAG, "Failed to initialize SPIFFS (%s)", esp_err_to_name(ret));
        }
        return;
    
    size_t total = 0, used = 0;
    ret = esp_spiffs_info(NULL, &total, &used);
    if (ret != ESP_OK)
        ESP_LOGE(TAG, "Failed to get SPIFFS partition information (%s)", esp_err_to_name(ret));
    else
        ESP_LOGI(TAG, "Partition size: total: %d, used: %d", total, used);
}

// init ports
void init_ports (void)
{
    int err;
    //sets GPIO15/16 as output ports for LED
    gpio_config_t io_conf;
    //disable interrupt
    io_conf.intr_type = GPIO_INTR_DISABLE;
    //set as output mode
    io_conf.mode = GPIO_MODE_OUTPUT;
    //bit mask of the pins that you want to set,e.g.GPIO15/16
    io_conf.pin_bit_mask = GPIO_OUTPUT_PIN_SEL;
    //disable pull-down mode
    io_conf.pull_down_en = 0;
    //disable pull-up mode
    io_conf.pull_up_en = 0;
    //configure GPIO with the given settings
    gpio_config(&io_conf);    
    gpio_set_level(GPIO_OUTPUT_IO_0, 0); 
    gpio_set_level(GPIO_OUTPUT_IO_1, 0);
    
    //interrupt of rising edge
    io_conf.intr_type = GPIO_INTR_NEGEDGE;
    //bit mask of the pins
    io_conf.pin_bit_mask = GPIO_INPUT_PIN_SEL;
    //set as input mode
    io_conf.mode = GPIO_MODE_INPUT;
    //enable pull-up mode
    io_conf.pull_up_en   = GPIO_PULLUP_ENABLE;
    io_conf.pull_down_en = GPIO_PULLDOWN_DISABLE;
    gpio_config(&io_conf);

    //change gpio intrrupt type for one pin

    ESP_LOGI(TAG, "install ISR service");
    err = gpio_install_isr_service(0);
    if (err != ESP_OK)
        ESP_LOGI(TAG, "install isr service failed with error 0x%x", err);
    //hook isr handler for specific gpio pin
    ESP_LOGI(TAG, "add isr handler");
    err = gpio_isr_handler_add(GPIO_INPUT_IO_0, &gpio_isr_handler, (void*) GPIO_INPUT_IO_0);
    if (err != ESP_OK)
        ESP_LOGI(TAG, "add isr handler add failed with error 0x%x", err);
}

/* init sntp */
static void init_sntp(void)
{
    ESP_LOGI(TAG, "Initializing SNTP");
    sntp_setoperatingmode(SNTP_OPMODE_POLL);
    sntp_setservername(0, NTPSERVICE);
    sntp_init();
} 

/* init WIFI */
void wifi_init_sta(void)
{
    s_wifi_event_group = xEventGroupCreate();
    //esp_netif_t interface;
    esp_netif_init();

    tcpip_adapter_ip_info_t ip_info;
    tcpip_adapter_dns_info_t dns_info;

    //set ip data if we use static IP
    if (strcmp(config.use_dhcp, "no") == 0 && gpio_get_level(GPIO_INPUT_IO_0))
        {
        tcpip_adapter_dhcpc_stop(TCPIP_ADAPTER_IF_STA);
        ip4addr_aton (config.ip, &ip_info.ip);
        ip4addr_aton (config.gw, &ip_info.gw);
        ip4addr_aton (config.netmask, &ip_info.netmask);
        tcpip_adapter_set_ip_info(TCPIP_ADAPTER_IF_STA, &ip_info);
        ip4addr_aton (config.dns1, &dns_info.ip);
        tcpip_adapter_set_dns_info(TCPIP_ADAPTER_IF_STA, TCPIP_ADAPTER_DNS_MAIN, &dns_info);
        ip4addr_aton (config.dns2, &dns_info.ip);
        tcpip_adapter_set_dns_info(TCPIP_ADAPTER_IF_STA, TCPIP_ADAPTER_DNS_BACKUP, &dns_info);
        }
    
        ESP_ERROR_CHECK(esp_event_loop_create_default());

        wifi_init_config_t cfg = WIFI_INIT_CONFIG_DEFAULT();
        ESP_ERROR_CHECK(esp_wifi_init(&cfg));
        
        ESP_ERROR_CHECK(esp_event_handler_register(WIFI_EVENT, ESP_EVENT_ANY_ID, &event_handler, NULL));
        ESP_ERROR_CHECK(esp_event_handler_register(IP_EVENT, IP_EVENT_STA_GOT_IP, &event_handler, NULL));

        
    wifi_config_t wifi_config = {
        .sta = {
            .ssid = "",
            .password = "",
        },
    };
 
    if (gpio_get_level(GPIO_INPUT_IO_0)) 
        {
        ESP_LOGI(TAG, "Button not pressed, using configured values for SSID, Wlan password, IP config and station password.");
        strcpy ((char *)wifi_config.sta.ssid, (char *)config.ssid);
        strcpy ((char *)wifi_config.sta.password, (char *)config.password);
        tcpip_adapter_set_hostname(TCPIP_ADAPTER_IF_STA, config.name);
        }
    else
        {
        ESP_LOGI(TAG, "Button pressed, using default values for SSID, Wlan password, IP config and station password.");
        strcpy ((char *)wifi_config.sta.ssid, WEBSERVER_ESP_WIFI_SSID);
        strcpy ((char *)wifi_config.sta.password, WEBSERVER_ESP_WIFI_PASS);
        strcpy (config.station_password, WEBSERVER_ESP_STATION_PASSWORD);
        tcpip_adapter_set_hostname(TCPIP_ADAPTER_IF_STA, WEBSERVER_ESP_STATION_NAME);
        }

    /* Setting a password implies station will connect to all security modes including WEP/WPA.
        * However these modes are deprecated and not advisable to be used. Incase your Access point
        * doesn't support WPA2, these mode can be enabled by commenting below line */

    if (strlen((char *)wifi_config.sta.password))
        wifi_config.sta.threshold.authmode = WIFI_AUTH_WPA2_PSK;

    ESP_ERROR_CHECK(esp_wifi_set_mode(WIFI_MODE_STA) );
    ESP_ERROR_CHECK(esp_wifi_set_config(ESP_IF_WIFI_STA, &wifi_config) );
    ESP_ERROR_CHECK(esp_wifi_start() );

    ESP_LOGI(TAG, "wifi_init finished.");

    /* Waiting until either the connection is established (WIFI_CONNECTED_BIT) or connection failed for the maximum
     * number of re-tries (WIFI_FAIL_BIT). The bits are set by event_handler() (see above) */
    EventBits_t bits = xEventGroupWaitBits(s_wifi_event_group,
            WIFI_CONNECTED_BIT | WIFI_FAIL_BIT,
            pdFALSE,
            pdFALSE,
            portMAX_DELAY);

    /* xEventGroupWaitBits() returns the bits before the call returned, hence we can test which event actually
     * happened. */
    if (bits & WIFI_CONNECTED_BIT)
        {
        ESP_LOGI(TAG, "connected to ap SSID:%s", wifi_config.sta.ssid);
        wifi_state = 1;
        //read out ip data in case of DHCP
        if (strcmp(config.use_dhcp, "yes") == 0)
            {
            //just to have the actual network data, we will store them within the config
            tcpip_adapter_get_ip_info(TCPIP_ADAPTER_IF_STA, &ip_info);
            tcpip_adapter_get_dns_info(TCPIP_ADAPTER_IF_STA, TCPIP_ADAPTER_DNS_MAIN, &dns_info);
            ip4addr_ntoa_r(&ip_info.ip,config.ip, sizeof(config.ip));
            ip4addr_ntoa_r(&ip_info.gw,config.gw, sizeof(config.gw));
            ip4addr_ntoa_r(&ip_info.netmask,config.netmask, sizeof(config.netmask));
            ip4addr_ntoa_r(&dns_info.ip,config.dns1, sizeof(config.dns1));
            tcpip_adapter_get_dns_info(TCPIP_ADAPTER_IF_STA, TCPIP_ADAPTER_DNS_BACKUP, &dns_info);
            ip4addr_ntoa_r(&dns_info.ip,config.dns2, sizeof(config.dns2));
            }
        }
    else if (bits & WIFI_FAIL_BIT)
        {
        ESP_LOGI(TAG, "Failed to connect to SSID:%s, password:%s",wifi_config.sta.ssid, wifi_config.sta.password);
        wifi_state = 0;
        }
    else
        {
        ESP_LOGE(TAG, "UNEXPECTED EVENT");
        wifi_state = 0;
        esp_restart();
        }

    ESP_ERROR_CHECK(esp_event_handler_unregister(IP_EVENT, IP_EVENT_STA_GOT_IP, &event_handler));
    ESP_ERROR_CHECK(esp_event_handler_unregister(WIFI_EVENT, ESP_EVENT_ANY_ID, &event_handler));
    vEventGroupDelete(s_wifi_event_group);
}

